@@ -5,7 +5,7 @@ use Mojo::Base 'Mojolicious::Plugin';
 use Mojo::Server::PSGI;
 use Plack::Util;
 use Mojo::Message::Request;
-our $VERSION = '0.17';
+our $VERSION = '0.19';
 
     ### ---
     ### register
@@ -18,7 +18,7 @@ our $VERSION = '0.17';
         my $plack_app = sub {
             my $env = shift;
             my $c = $env->{'MOJO.CONTROLLER'};
-            $c->req(Mojo::Message::Request->new->parse($env));
+            $c->tx->req(psgi_env_to_mojo_req($env));
             $on_process_org->($c->app, $c);
             return mojo_res_to_psgi_res($c->res);
         };
@@ -53,6 +53,27 @@ our $VERSION = '0.17';
         });
     }
     
+    use constant CHUNK_SIZE => $ENV{MOJO_CHUNK_SIZE} || 131072;
+    
+    sub psgi_env_to_mojo_req {
+        
+        my $env = shift;
+        my $req = Mojo::Message::Request->new->parse($env);
+        
+        # Request body
+        my $len = $env->{CONTENT_LENGTH};
+        while (!$req->is_done) {
+            my $chunk = ($len && $len < CHUNK_SIZE) ? $len : CHUNK_SIZE;
+            my $read = $env->{'psgi.input'}->read(my $buffer, $chunk, 0);
+            last unless $read;
+            $req->parse($buffer);
+            $len -= $read;
+            last if $len <= 0;
+        }
+        
+        return $req;
+    }
+    
     ### ---
     ### convert mojo tx to psgi env
     ### ---
@@ -62,16 +83,23 @@ our $VERSION = '0.17';
         my $url = $mojo_req->url;
         my $base = $url->base;
         my $host = $base->host;
+        my $length = length($mojo_req->body);
+        my $body = Mojolicious::Plugin::PlackMiddleware::PSGIInput->new($mojo_req->body);
+        
+        my %headers = %{$mojo_req->headers->to_hash};
+        for my $key (keys %headers) {
+           my $value = $headers{$key};
+           delete $headers{$key};
+           $key =~ s{-}{_};
+           $headers{'HTTP_'. uc $key} = $value;
+        }
+        
         return {
             %ENV,
+            %headers,
             'SERVER_PROTOCOL'   => 'HTTP/'. $mojo_req->version,
             'SERVER_NAME'       => $host,
             'SERVER_PORT'       => $base->port,
-            'HTTP_HOST'         => $host. ':' .$base->port,
-            'HTTP_AUTHORIZATION' => sub {
-                    my $a = $mojo_req->headers->header('authorization');
-                    return $a || '';
-                }->(),
             'REQUEST_METHOD'    => $mojo_req->method,
             'SCRIPT_NAME'       => '',
             'PATH_INFO'         => $url->path->to_string,
@@ -81,6 +109,8 @@ our $VERSION = '0.17';
             'psgi.multithread'  => Plack::Util::FALSE,
             'psgi.version'      => [1,1],
             'psgi.errors'       => *STDERR,
+            'CONTENT_LENGTH'    => $length,
+            'psgi.input'        => $body,
             'psgi.multithread'  => Plack::Util::FALSE,
             'psgi.multiprocess' => Plack::Util::TRUE,
             'psgi.run_once'     => Plack::Util::FALSE,
@@ -186,6 +216,23 @@ use parent qw(Plack::Middleware::Conditional);
             return $self->app->($env);
         }
     }
+    
+package Mojolicious::Plugin::PlackMiddleware::PSGIInput;
+use strict;
+use warnings;
+    
+    sub new {
+        my ($class, $content) = @_;
+        return bless [$content, 0], $class;
+    }
+    
+    sub read {
+        my $self = shift;
+        if ($_[0] = substr($self->[0], $self->[1], $_[1])) {
+            $self->[1] += $_[1];
+            return 1;
+        }
+    }
 
 1;
 
@@ -267,6 +314,12 @@ $plugin->register;
 
 Register plugin hooks in L<Mojolicious> application.
 
+=head2 psgi_env_to_mojo_req
+
+This is a utility method. This is for internal use.
+
+    my $mojo_req = psgi_env_to_mojo_req($psgi_env)
+
 =head2 mojo_req_to_psgi_env
 
 This is a utility method. This is for internal use.
